;;; The EpiShell command hash package.  Provides hash table lookup of commands,
;;; their synonyms, and abbreviations.

(in-package user)
(eval-when (compile load eval)   
(provide 'esh-cmdhash)
(defpackage esh-cmdhash 
  (:use lisp #+lucid lucid-common-lisp #-lucid common-lisp #+excl excl)
  (:export
          make-cmd-hash-table  	   ;make or extend a command hash table,
          get-cmd-hash	  	   ;lookup a command in a command hash table
	  option-p		   ;true if a given symbol is an option (i.e.,
				   ; starts with a dash)
	 ))
(unless (find-package 'esh-cmdhash) (make-package 'esh-cmdhash)))
(in-package esh-cmdhash)



; Creates or adds entries to a command hash table.
; Each command or synonym and its their abbreviations are added to the
; table as separate entries.
; To keep the size of the table down, the keys are truncated to be
; just long enough to differentiate between commands.
; Commands and synonyms should be symbols.
; Each entry in the table is a list of the form: 
;        (non-abbrev-flag command-or-synonym-name . command-name)

(defun make-cmd-hash-table  (cmd-list &optional 
             (hash-table (make-hash-table :test #'equal
					  :size (* 5 (length cmd-list))))
	     &aux (key-length (or (gethash "KeYlEnGtH" hash-table) 
				  (setf (gethash "KeYlEnGtH" hash-table) 2))))

    (mapcar #'(lambda (cmd)
		(mapcar #'(lambda (synonym &aux syn entry)
			(setq synonym (get-cmd-name synonym))
			(do nil
			    (nil)
			    (setq syn (cmd-truncate synonym key-length))
			    (setq entry (gethash syn hash-table))
			    (if (or (not entry) (string= (cadr entry) synonym)
				    (> key-length (length synonym)))
				(return))
			    (setq key-length 
                                            (rehash-cmd-table hash-table)))
                        (setf (gethash syn hash-table)   
				           (cons t (cons synonym (car cmd))))
			(mapcar #'(lambda (abbrev)
				(unless (gethash abbrev hash-table)
					(setf (gethash abbrev hash-table)
					      (cons nil 
						   (cons synonym (car cmd))))))
				(get-abbrevs syn)))
			cmd))
	 cmd-list)
   hash-table
)
  
;;; Changes a command name or synonym to a string. Strips off the
;;; package name.

(defun get-cmd-name (name &aux colon)
    (setq name (prin1-to-string name))
    (if (setq colon (position #\: name :from-end t))
	(subseq name (+ 1 colon))
	name) 
)


;;; Increment the keylength for the table by 1, and then
;;; go through the whole table, creating new keylength+1 entries for
;;; the non-abbrev entries in the table.

(defun rehash-cmd-table (hash-table 
		&aux (key-length (incf (gethash "KeYlEnGtH" hash-table)))
		     (new-entries nil))
   (maphash #'(lambda (key value)
	(if (and (listp value) (car value) (string/= key (cadr value)))
	    (push (cons key (copy-list value)) new-entries)))
        hash-table)
   (mapc #'(lambda (entry)
                ;; Mark the old entry as an abbreviation
                (setf (car (gethash (car entry) hash-table)) nil)
		(setf (gethash (cmd-truncate (caddr entry) key-length) 
			       hash-table) (cdr entry)))
	 new-entries)
   key-length)

(defun cmd-truncate (str len)
    (if (> len (length str))
	str
        (subseq str 0 len)))
     
;;; Returns all the proper prefixes of length >= 2 of a string.
;;;
(defun get-abbrevs (str)
    (do ((char-list 
	      (cdr (nreverse (map 'list #'char-code str)))
	      (cdr char-list))
	 (abbrevs nil))
	((or (not char-list) (not (cdr char-list))) abbrevs)
	(setq abbrevs 
		(cons (map 'string #'character (reverse char-list))
		      abbrevs)))
)


;;; Truncate the cmd to the keylength of the table and
;;; use it hash to an entry in the table.
;;; If an entry is found, make sure the truncation didn't
;;; give us a false match.

(defun get-cmd-hash (cmd hash-table &aux hash-entry)
    (if (symbolp cmd)
        (if (setq hash-entry 
		  (gethash (cmd-truncate (setq cmd (prin1-to-string cmd))
		                            (gethash "KeYlEnGtH" hash-table))
	                   hash-table))
	    (if (string= (cadr hash-entry) cmd 
				:end1 (min (length (cadr hash-entry))
					   (length cmd)))
		(cddr hash-entry))))
)


(defun option-p (sym)
    (and (symbolp sym) (eql (elt (prin1-to-string sym) 0) #\-)))
