;       Last change    August 1993
(in-package time)
;        Routines to get and create Time nets, points and events

; Short cuts to find out link information

(defmacro from-chain-number (lk)
  `(meta-node-chain-number
    (time-node-chain
     (time-point
      (time-link-from-point-name ,lk)))))

(defmacro from-psuedo (lk)
  `(time-node-psuedo (time-point (time-link-from-point-name ,lk))))

(defmacro from-name (lk)
  `(time-node-name (time-point (time-link-from-point-name ,lk))))

(defmacro to-chain-number (lk)
  `(meta-node-chain-number
    (time-node-chain
     (time-point
      (time-link-to-point-name ,lk)))))

(defmacro to-psuedo (lk)
  `(time-node-psuedo (time-point (time-link-to-point-name ,lk))))

(defmacro to-name (lk)
  `(time-node-name (time-point (time-link-to-point-name ,lk))))

; On-same-chain returns t if the two points are on the same chain, nil otherwise
(defmacro on-same-chain (t1 t2)
  `(equal (time-node-chain ,t1)
	  (time-node-chain ,t2)))

; First-in-chain returns t if the point is the first on its chain - i.e. it has
; no ancestors on the same chain

(defmacro first-in-chain (pt)
  `(null (time-node-ancestors ,pt))
  )

; Last-in-chain returns t if the point is last in its chain - i.e. it has no
; descendants on the same chain

(defmacro last-in-chain (pt)
  `(null (time-node-descendants ,pt))
  )



; Test-adjust checks to see if there is enough room in the array to add
; another entry, and if not, adjusts it so that there is
; this is used to expand the metagraph if necessary

(defun test-adjust (graph entry &aux
			  (num (car (array-dimensions graph)))
			  (array graph))
  (cond ((>= entry num)
	 (setq array (adjust-array array (+ num '5)
				   :initial-element nil))
	 (test-adjust array entry))
	(t array)
	)
  )

(defun first-from-tree (func &optional (env *environment*))
  "Apply func to env and its descendants, in a
  preorder, and return the first non-nil answer that is found."
  (let ((result (funcall func env)))
    (if result
	result
      (dolist (e (environment-use-environments env))
	(setq result (first-from-tree func e))
	(if result
	    (return result))))))


; writeable-net returns a time net for context *current-netname* in
; environment *environment*, creating it if necessary.

(defun writeable-net ()
  (let ((context-array (retrieve-specialist-info
			*environment*
			'time
			(lambda ()
			  (make-hash-table :test 'eql
					   :size '100
					   :rehash-size '10
					   :rehash-threshold '.8)))))
    (when (null (geth *current-netname* context-array))
      (seth *current-netname* context-array (newnet *current-netname*)))
    (geth *current-netname* context-array)))


;; writeable-point takes either a time-node or a name (concept).  If
;; it is a name, it first retrieves the corresponding time-node.
;; Then, if the time-node is in the current environment, it returns
;; it, otherwise it makes a copy in the current environment and
;; returns that.  Currently, it also copies the entire chain that the
;; node belongs to.  That might change in the future.

(defun writeable-point (name-or-point &aux name point)
  (assert name-or-point)
  (if (time-node-p name-or-point)
      (progn
	(setq point name-or-point)
	(setq name (time-node-name point)))
    (progn
      (setq name name-or-point)
      (setq point (time-point name))))
  (if (eq (time-node-environment point)
	  *environment*)
      point
    (progn
      (writeable-chain (meta-node-chain-number (time-node-chain point)))
      ;; ignore return value, just use for side-effect of copying chain
      (geth name (time-net-time-graph (writeable-net))))))

(defun writeable-chain (number)
  (let ((oldchain (time-chain number)))
    (if (eq (meta-node-environment oldchain)
	    *environment*)
	oldchain
      (let ((net (writeable-net))
	    (newchain (make-meta-node
		       :chain-number number
		       :connections (meta-node-connections oldchain))))
	(setf (meta-node-first newchain)
	  (copy-chain-nodes (meta-node-first oldchain)
			    newchain
			    net))
	(change-hash number
		     (time-net-meta-graph net)
		     newchain)))))

(defun writeable-link (link)
  (if (writeable-p link)
      link
    (progn
      (writeable-point (time-link-from-point-name link))
      (writeable-point (time-link-to-point-name link))
      (copy-time-link link))))

(defun writeable-p (obj)
  (cond ((time-node-p obj) (eq *environment* (time-node-environment obj)))
	((meta-node-p obj) (eq *environment* (meta-node-environment obj)))
	((time-link-p obj) (and (writeable-p
				 (time-point
				  (time-link-from-point-name obj)))
				(writeable-p
				 (time-point
				  (time-link-to-point-name obj)))))
	((time-net-p obj) (eq *environment* (time-net-environment obj)))
	((event-node-p obj) (eq *environment* (event-node-environment obj)))
	(t (print-error "writeable-p called with argument " obj "of unrecognized type.")
	   nil)))


;; copy-chain-nodes copies all nodes of a chain, and enters the copies
;; into the supplied net.  It returns the first node in the copied
;; chain.

(defun copy-chain-nodes (oldnode newmetanode net)
  ;; Note that the ancestor and descendant lists from the old
  ;; structure are reused.  Therefore, those lists must not be
  ;; modified destructively.
  (let ((newnode (copy-time-node oldnode)))
    (setf (time-node-chain newnode) newmetanode)
    (setf (time-node-environment newnode) *environment*)
    (change-hash (time-node-name newnode)
		 (time-net-time-graph net)
		 newnode)
    (if (time-node-descendants oldnode)
	(copy-chain-nodes (time-point (time-link-to-point-name
				       (car (time-node-descendants oldnode))))
			  ;; only necessary to copy the first descendant
			  ;; in the list, because a chain is a total order
			  ;; and the descendants list is ordered by
			  ;; pseudo-time.
			  newmetanode
			  net))
    newnode))


; Time-point returns the time point corresponding to ptname if there
; is one. If not, nil is returned.

(defun time-point (ptname)
  (case ptname
    ((nil t) ptname)
    (otherwise (first-from-tree
		#'(lambda (env)
		    (let ((context-array (retrieve-specialist-info env 'time)))
		      (if context-array
			  (let ((net (geth *current-netname* context-array)))
			    (if net
				(geth ptname (time-net-time-graph net)))))))))))

; Time-chain returns the meta-node corresponding to chainno if there
; is one. If not, nil is returned.

(defun time-chain (chainno)
  (first-from-tree
   #'(lambda (env)
       (let ((context-array (retrieve-specialist-info env 'time)))
	 (if context-array
	     (let ((net (geth *current-netname* context-array)))
	       (if net
		   (geth chainno (time-net-meta-graph net)))))))))

; Newnet creates a new time net, including the time-point table, metagraph
; array (for chains), and an event table.

(defun newnet (ident &aux

		     ; hash table for time points

		     (nt (make-hash-table :test 'eql
					  :size '100
					  :rehash-size '10
					  :rehash-threshold '.8)) ; changed!
                                             ;; threshold value changed from 9
                                             ;; [2/24/95 CH]

		     ; array for chains (all numeric)

		     (nm (make-hash-table :test '=
					  :size 10
					  :rehash-size 10
					  :rehash-threshold .8))

		     ; hash table for events

		     (ne (make-hash-table :test 'eql
					  :size '50
					  :rehash-size '10
					  :rehash-threshold '.8))) ; changed!
                                        ;; threshold value changed from 9
                                        ;; [2/24/95 CH]

  ; put it all together in one time net node

  (make-time-net :identifier ident
		 :time-graph nt
		 :meta-graph nm
		 :event-table ne)
  )

; event-point returns the event point corresponding to pt if one exists

(defun event-point (pt)
  (first-from-tree
   #'(lambda (env)
       (let ((context-array (retrieve-specialist-info env 'time)))
	 (if context-array
	     (let ((net (geth *current-netname* context-array)))
	       (if net
		   (geth pt (time-net-event-table net)))))))))


; prepare-event sets up an event node with its start and end points, if
; one does not already exist

(defun prepare-event (eventname &aux
			  (e (event-point eventname)))
  (when (null e)
    (let ((ename (find-symbol (symbol-name eventname) 'epilog)))
      (setq e (change-hash
	       eventname
	       (time-net-event-table (writeable-net))
	       (make-event-node :name eventname
				:start (pack (list ename 'epilog::start))
				:end (pack (list ename 'epilog::end)))))))
  e)

; absolute-time returns t if arg is an absolute time, nil otherwise
; absolute times are quoted expressions of the form (quote (time yy mm
; dd hh min ss))

(defun absolute-time (arg)
  (and (symbolp arg) (equal (getp arg 'term-type) 'record)))

; get-abs gets an actual absolute time representation from an absolute time
; concept

(defun get-abs (arg)
  (mapcar #'(lambda (a) (if (numberp a) (round a) a))
	  (record-contents arg))
  )

; event returns t if arg is a concept of sort event/episode/interval

(defun event (arg)
  (if (symbolp arg)
      (let ((sort (find-symbol (symbol-name (find-sort arg)) 'time))) ;get-sort
           (member sort '(episode event interval))  
	   )
      nil)
  )

; get-start returns the name of the start point of an event if the given
; concept is an event. Otherwise it returns the concepts name

(defun get-start (concept)
  (if (null concept)
      nil
      (if (event concept)
	  (event-node-start (prepare-event concept))
	  concept ; time point
	  )
      )
  )

; get-end returns the name of the end point of an event if the given
; concept is an event. Otherwise it returns the concepts name

(defun get-end (concept)
  (if (null concept)
      nil
      (if (event concept)
	  (event-node-end (prepare-event concept))
	  concept ; time point
	  )
      )
  )

; abs-start returns the start of an argument if it is a concept, and the
; absolute time itself if it is an absolute time

(defmacro abs-start (arg)
  `(if (absolute-time ,arg)
       ,arg
       (get-start ,arg))
  )

; abs-end returns the end of an argument if it is a concept, and the
; absolute time itself if it is an absolute time

(defmacro abs-end (arg)
  `(if (absolute-time ,arg)
       ,arg
       (get-end ,arg))
  )

(defun missing (arg)
  (if (event arg)
      (not (event-point arg))
      (not (time-point arg))
      )
  )

; max-num returns the maximum of two items. If either is not a number there is
; a problem

(defun max-num (n1 n2)
  (if (numberp n1)
      (if (numberp n2)
          (max n1 n2)
	;; who knows for other case? need a number specialist
	n2)
    (if (numberp n2)
	n2
      n2)
    )
  )
 
; min-num returns the minimum of two items. If either is not a number there is
; a problem
 
(defun min-num (n1 n2)
  (if (numberp n1)
      (if (numberp n2)
          (min n1 n2)
          ; who knows for other case? need a number specialist
          n2)
      (if (numberp n2)
          n2
          n2)
      )     
  )

