;       Last change    August 1993
(in-package time)
;                         LINKS
;
; A link connects two time points together. This link contains pointers to the
; starting time point and the ending time point, as well as strictness and 
; duration information.
; A link is stored on the descendant (or xdescendant) list of the starting
; point, as well as the ascendant (or xascendant) list of the end point. In
; addition, it the points are on different chains, the link is also stored
; on the connection list for the chain of the start point.
;
; This section contains routines for adding links and maintaining the lists,
; as well as for removing and copying the links when two points are made equal.

; Short cuts to find out link information

; Test-insert returns t if item should be inserted before the first link 
; in llist, nil otherwise
; The lists of links are ordered from chain, from psuedo, to chain, to psuedo

(defun test-insert (llist item &aux
			  (lk (car llist)))
  (cond ((or (> (from-chain-number lk) (from-chain-number item))
	     (and (= (from-chain-number lk) (from-chain-number item))
		  (or (> (from-psuedo lk) (from-psuedo item))
		      (and (= (from-psuedo lk) (from-psuedo item))
			   (or (> (to-chain-number lk) (to-chain-number item))
			       (and (= (to-chain-number lk) (to-chain-number item))
				    (>= (to-psuedo lk) (to-psuedo item))
				    )
			       )
			   )
		      )
		  )
	     )
                 
	 t)
        (t nil))
  )

; Ins-here inserts item at the beginning of llist. If item is already there,
; it is not added. If it is there but the strictness is different, the most
; strict value is used for the link.

(defun ins-here (llist item &aux
		       (lk (car llist)))

  ; see if item is already here

  (cond ((and (equal (from-chain-number lk) (from-chain-number item))
              (equal (from-psuedo lk) (from-psuedo item))
              (equal (to-chain-number lk) (to-chain-number item))
              (equal (to-psuedo lk) (to-psuedo item)))

	 ; if the new strictness is strict, change the strictness of the link
	 ; otherwise there is no need to change it

	 (when (time-link-strict item) 
	       (change-field 'time-link-strict lk t)
	       )

	 ; return the list itself

	 llist)

	; otherwise put item at the front of the list

        (t (append (list item) llist)))
  )

; Insinlist inserts item at the appropriate place in llist, and returns the new
; list

(defun insinlist (llist item)

  ; if the list is null, make a new one consisting of item

 (cond ((null llist) (list item))

       ; check to see if item goes before list

       ((test-insert llist item)
        (ins-here llist item))

       ; insert item somewhere in the rest of the list

       (t (append (list (car llist)) (insinlist (cdr llist) item))))
  )

; Add-meta-link adds a link to the meta graph for the appropriate chain
; These are thie links actually used in searching the graph.

(defun add-meta-link (timelink)
  (unless (equal (from-chain-number timelink) (to-chain-number timelink))
    (let ((mn (writeable-chain (from-chain-number timelink))))
      (change-field 'meta-node-connections mn
		    (insinlist (meta-node-connections mn) timelink))
      )
    )
  )

; Add-ancestor-link adds a link on the in chain ancestor list for point
; This list is ordered by from chain and psuedo and to chain and psuedo

(defun add-ancestor-link (point timelink)
  (change-field 'time-node-ancestors point
          (insinlist  (time-node-ancestors point) timelink))
  )

; Add-decendant-link adds a link on the in chain dewcendant list for point
; This list is ordered by from chain and psuedo and to chain and psuedo

(defun add-descendant-link (point timelink)
  (change-field 'time-node-descendants point
          (insinlist  (time-node-descendants point) timelink))
  )

; Add-xancestor-link adds a link on the cross chain ancestor list for point
; This list is ordered by from chain and psuedo and to chain and psuedo

(defun add-xancestor-link (point timelink)
  (change-field 'time-node-xancestors point
          (insinlist  (time-node-xancestors point) timelink))
  )

; Add-xdescendant-link adds a link on the cross chain descendant list for point
; This list is ordered by from chain and psuedo and to chain and psuedo

(defun add-xdescendant-link (point timelink)
  (change-field 'time-node-xdescendants point
          (insinlist  (time-node-xdescendants point) timelink))
  )

; Add-link adds a link between pt1 and pt2 with the appropriate strictness
; If the two point are on different chains, a meta link is also added
; Links are added to the ancestor (or xancestor) list for pt2 and
; the descendant or (xdescendant) list for pt1

(defun add-link (pt1 pt2 strict12 &aux tl)

  (assert (writeable-p pt1))
  (assert (writeable-p pt2))
  ; if the two points are equal, do not add a link between them!!!

  (unless (and (equal (time-node-chain pt1)
		      (time-node-chain pt2))
	       (equal (time-node-psuedo pt1)
		      (time-node-psuedo pt2)))
	  (setq tl (make-time-link :from-point-name (time-node-name pt1)
				   :to-point-name (time-node-name pt2)
				   :strict (strict strict12)))

          ; add to descendant or xdescendant list of pt1
          ; add to ancestor or xancestor list of pt2
          ; if x link, add to metagraph

          (cond ((equal (time-node-chain pt1) (time-node-chain pt2))
                 (add-descendant-link pt1 tl)
                 (add-ancestor-link pt2 tl))
                (t (add-xdescendant-link pt1 tl)
                   (add-xancestor-link pt2 tl)
                   (add-meta-link tl)))
	  tl
	  )
  )

; The following routines are used when copying information from one node
; to another

; Same-link returns t if l1 is the same link as l2

(defun same-link (l1 l2)
  (cond ((and (equal (from-chain-number l1) (from-chain-number l2))
              (equal (from-psuedo l1) (from-psuedo l2))
              (equal (to-chain-number l1) (to-chain-number l2))
              (equal (to-psuedo l1) (to-psuedo l2)))
         t)
        (t nil))
  )

; Remove-item removes item from llist, using same-link to determine whether
; it has the correct link
; The new list minus item is returneds

(defun remove-item (llist item)
  (cond ((null llist) nil)
	((not (listp llist))
	 (if (same-link llist item)
	     nil
	     llist))
        ((same-link (car llist) item)
         (cdr llist))
        (t (append (list (car llist)) (remove-item (cdr llist) item))))
  )

; Remove-meta-link removes a link from the meta graph link for the appropriate
; chain (the from chain)

(defun remove-meta-link (timelink)
  (unless (equal (from-chain-number timelink) (to-chain-number timelink))
          (let ((mn (writeable-chain (from-chain-number timelink))))
	    (change-field 'meta-node-connections mn
			  (remove-item (meta-node-connections mn) timelink))
	    )
	  )
  )

; Remove-link removes item from linklist, and also removes a meta-link if there
; is one for it.

(defun remove-link (item linklist)
  (remove-meta-link item)
  (remove-item linklist item)
)

; Get-desc-list returns a list of all the descendants of a point
; x indicates whether to return the in chain or cross chain descendants

(defun get-desc-list (point x)
  (if (null x)
      (time-node-descendants point)
      (time-node-xdescendants point)
      )
  )

; Update-descendants  copies descendants -again x indicates whether in chain
; or x chain - from pt1 to pt2. For each link on pt1's descendant list, the
; "to" point ancestor list has this link removed, and then the link is
; added using pt2 as the ancestor.

(defun update-descendants (pt1 pt2 x)
  (assert (writeable-p pt1))
  (assert (writeable-p pt2))
  (dolist (item (get-desc-list pt1 x))
          (let ((pt (writeable-point (time-link-to-point-name item)))
		(durmin (time-link-duration-min item))
		(durmax (time-link-duration-max item)))
               (remove-ancestor pt item x)
               (add-link pt2 pt (if (time-link-strict item) '1 nil))
               ; add duration as well if there was any
	       (unless (null durmin)
		       (new-duration-min pt2 pt durmin))
	       (unless (null durmax)
		       (new-duration-max pt2 pt durmax))
               )
          )
  )

; Remove descendant removes an item from the descendant list for point
; x indicates whether to remove from the in chain or cross chain descendants

(defun remove-descendant (point item x)
  (if (null x)
      (change-field 'time-node-descendants point
              (remove-link item (time-node-descendants point)))
      (change-field 'time-node-xdescendants point
              (remove-link item (time-node-xdescendants point)))
       )
  )

; Get-anc-list returns a list of all the ancestors of a point
; x indicates whether to return the in chain or cross chain ancestors

(defun get-anc-list (point x)
  (if (null x)
      (time-node-ancestors point)
      (time-node-xancestors point)
      )
  )

; Update-ancestors  copies ancestors -again x indicates whether in chain
; or x chain - from pt1 to pt2. For each link on pt1's ancestor list, the
; "from" point descendant list has this link removed, and then the link is
; added using pt2 as the descendant.

(defun update-ancestors (pt1 pt2 x)
  (assert (writeable-p pt1))
  (assert (writeable-p pt2))
  (dolist (item (get-anc-list pt1 x))
          (let ((pt (writeable-point (time-link-from-point-name item)))
		(durmin (time-link-duration-min item))
		(durmax (time-link-duration-max item)))
               (remove-descendant pt item x)
               (add-link pt pt2 (if (time-link-strict item) '1 nil))
               ; add duration as well if there was any
	       (unless (null durmin)
		       (new-duration-min pt pt2 durmin))
	       (unless (null durmax)
		       (new-duration-max pt pt2 durmax))
               )
          )
  )

; Remove ancestor removes an item from the ancestor list for point
; x indicates whether to remove from the in chain or cross chain ancestor

(defun remove-ancestor (point item x)
  (if (null x)
      (change-field 'time-node-ancestors point
              (remove-link item (time-node-ancestors point)))
      (change-field 'time-node-xancestors point
              (remove-link item (time-node-xancestors point)))
      )
  )

; Copy links copies all links (x chain and in chain) for pt1 to pt2, ensuring
; that the only links with points on the same chain go into the new
; in chain lists, and only those with different chains go on the x chain lists.

(defun copy-links (pt1 pt2)

  ; if points on same chain, add links from ancestors and descendants of
  ; pt1 to the corresponding lists for pt2, and also do xanc and xdesc
  ; if on different chains, anc and desc go on xanc and xdesc of pt2,
  ; and each of xanc and xdesc has to be checked to see if chain is equal

  (assert (writeable-p pt1))
  (assert (writeable-p pt2))
  (update-ancestors pt1 pt2 nil)
  (update-ancestors pt1 pt2 t)
  (update-descendants pt1 pt2 nil)
  (update-descendants pt1 pt2 t)
  )

